""" Init file """
import inspect

from .fields.base import Field
from .fields.boolean import BooleanField
from .fields.char import CharField
from .fields.email import EmailField
from .fields.id import IdField
from .fields.json import JsonField
from .fields.number import NumberField


class Form:
  """
  Form class
  ---
  Notes:
    - Any member that starts with `_` will be ignored.
  """

  _obj = {}
  _errors = {}
  _clean_functions = []
  _attributes = {}
  _nested_attrs = {}
  _sub_forms_attrs = {}

  def __init__(self, obj=dict):
    """ Constructor """
    self._obj = obj
    self.calculate_members()

  @property
  def cleaned_data(self):
    """ Returns the cleaned data """
    return self._obj

  def calculate_members(self):
    """ Calculate members """
    self._errors = {}
    self._clean_functions = []
    self._attributes = {}
    self._nested_attrs = {}
    self._sub_forms_attrs = {}

    for item in inspect.getmembers(self):
      if item[0] in self._reserverd_words:
        continue
      if item[0].startswith('_'):
        continue
      elif item[0].startswith('clean'):
        self._clean_functions.append(item[0])
      elif isinstance(item[1], Field):
        self._attributes[item[0]] = item[1]
      elif isinstance(item[1], list):
        self._nested_attrs[item[0]] = item
      elif isinstance(item[1], Form):
        self._sub_forms_attrs[item[0]] = item[1]
      else:
        print('Unknown field', item[0])

  def set_obj(self, obj):
    """ Set the object """
    self._obj = obj

  def is_valid(self):
    """ Returns if the form is valid """
    self._errors = {}

    for field in self._attributes.items():
      self._validate_field(field=field)

    for field, form in self._sub_forms_attrs.items():
      self._validate_sub_form(field=field, form=form, data=self._obj.get(field, {}))

    for field, form in self._nested_attrs.items():
      self._validate_sub_form_as_list(field=field, form=form[0])

    for func in self._clean_functions:
      self._clean(clean_func=func)

    return len(self._errors) == 0

  def errors(self):
    """ Returns the list of errors """
    return self._errors

  def add_errors(self, key='', code='', extra_args=dict):
    """ Add custom errors
    This function is designed to be used in a clean function
    ---
    Arguments
      key: str
        Key of the field
      code: str
        Code of the error
      extra_args: dict
        Extra arguments to add to the error
    """
    if key == '' or code == '':
      raise Exception('key and code are required')  #pylint: disable=W0719
    camel_key = self._convert_to_camel(key=key)

    if camel_key not in self._errors:
      self._errors[camel_key] = []

    new_error = {'code': code}
    if extra_args and isinstance(extra_args, dict):
      if callable(extra_args):
        extra_args = extra_args()

      new_error.update(extra_args)
    self._errors[camel_key].append(new_error)

  def _validate_field(self, field, new_key=None):
    """ Validate field """
    if isinstance(field[1], Field):
      func = getattr(field[1], 'validate')
      if callable(func):
        # Validate if the validate function has the correct parameters
        params = [p for p, _ in inspect.signature(func).parameters.items()]
        valid_params = ['key', 'value', 'errors']

        if len(params) != len(valid_params):
          raise Exception(f'{type(field[1])} validate method has no the correct parameters')  #pylint: disable=W0719

        is_valid = False
        for param in params:
          if param in valid_params:
            is_valid = True
            continue
          is_valid = False
          break

        if not is_valid:
          raise Exception(   #pylint: disable=W0719
            f'{field[0]} of type {type(field[1]).__name__} validate method has no the correct ' +\
            f'parameters. Expected parameters: {", ".join(valid_params)}. ' +\
            f'Actual parameters: {", ".join(params)}'
          )

        field[1].validate(
          key=field[0] if new_key is None else new_key,
          value=self._obj.get(field[0], None),
          errors=self._errors,
        )
      else:
        raise Exception(f'{type(field[1])} has no validate method')  #pylint: disable=W0719

  def _clean(self, clean_func):
    """ Clean function """
    func = getattr(self, clean_func)
    if callable(func):
      func()

  def _convert_to_camel(self, key):
    """
    Convert the key to camel case
    """
    init, *temp = key.split('_')

    field = ''.join([init, *map(str.title, temp)])
    field_items = field.split('.')

    field_final = []
    for item in field_items:
      field_final.append(''.join([item[0].lower(), item[1:]]))

    return '.'.join(field_final)

  def _validate_sub_form(self, field, form, data):
    """ Validate sub form """
    if not isinstance(form, Form):
      return
    form.set_obj(data)
    form.calculate_members()
    if not form.is_valid():
      for key, errors in form.errors().items():
        for error in errors:
          code = error['code']
          del error['code']
          self.add_errors(key=f'{field}.{key}', code=code, extra_args=error)

  def _validate_sub_form_as_list(self, field, form):
    """ Validate sub form for list """
    list_obj = self._obj.get(field, [])

    if isinstance(list_obj, (list, tuple)):
      for i, obj in enumerate(list_obj):
        if isinstance(form, Field):
          self._validate_field(
            field=obj,
            new_key=f'{field}.{i}',
          )
        elif isinstance(form, Form):
          self._validate_sub_form(
            field=f'{field}.{i}',
            form=form,
            data=obj,
          )

  @property
  def _reserverd_words(self):
    """ Reserved words """
    return (
      'add_errors',
      'change_obj',
      'clean',
      'errors',
      'is_valid',
      'set_obj',
      'calculate_members',
      'cleaned_data',
    )



""" Boolean field """
from .base import Field


class NumberField(Field):
  """
  NumberField class for validation
  ---
  Attributes
    required: bool
      Indicates if the field is required or not
    datatype: (int, float)
      Indicates the datatype of the field
    min_value: int
      Indicates the minimum value of the field
    max_value: int
      Indicates the maximum value of the field
  """

  def __init__(
    self,
    required=False,
    datatype=float,
    min_value=None,
    max_value=None,
  ):
    super(NumberField, self).__init__(required=required)
    self.datatype = datatype
    self.min_value = min_value
    self.max_value = max_value

  def validate(self, key, value, errors):
    """
    Validate the field with the following rules:
    - Should be a int or float (Depending of the datatype)
    ---
    Arguments
      key: str
        Key of the field
      value: any
        Value to validate
      errors: dict
        Dict of errors
    """

    super(NumberField, self).validate(key=key, value=value, errors=errors)

    if not isinstance(value, self.datatype) and (self.required and value is not None):
      self._append_error(key=key, errors=errors, to_add={'code': 'invalid'})
    else:
      try:
        if self.min_value is not None:
          if self.datatype(value) < self.datatype(self.min_value):
            self._append_error(
              key=key,
              errors=errors,
              to_add={
                'code': 'minValue',
                'expected': self.datatype(self.min_value),
                'received': self.datatype(value),
              },
            )
        if self.max_value is not None:
          if self.datatype(value) > self.datatype(self.max_value):
            self._append_error(
              key=key,
              errors=errors,
              to_add={
                'code': 'maxValue',
                'expected': self.datatype(self.max_value),
                'received': self.datatype(value),
              },
            )
      except ValueError:
        if self.required:
          self._append_error(
            key=key,
            errors=errors,
            to_add={'code': 'invalid'},
          )



""" Email field """
from .base import Field


class CharField(Field):
  """
  CharField class for validation
  ---
  Attributes
    required: bool
      Indicates if the field is required or not
    max_length: int
      Indicates the maximum length of the field
    min_length: int
      Indicates the minimum length of the field
    empty: bool
      Indicates if the field can be empty or not
    choices: tuple
      Indicates the choices of the field
  """

  def __init__(
    self,
    required=False,
    max_length=None,
    min_length=None,
    empty=False,
    choices=None,
  ):
    super(CharField, self).__init__(required=required)
    self.max_length = max_length
    self.min_length = min_length
    self.empty = empty
    self.choices = choices

  def validate(self, key, value, errors):
    """
    Validate the field with the following rules:
    - Should not be empty if required
    - Should be one of the choices indicated if choices is not None
    - Should be less than max_length if max_length is not None
    - Should be greater than min_length if min_length is not None
    ---
    Arguments
      key: str
        Key of the field
      value: any
        Value to validate
      errors: dict
        Dict of errors
    """

    super(CharField, self).validate(key=key, value=value, errors=errors)

    if value is not None:
      if not self.empty:
        if len(value) == 0:
          self._append_error(
            key=key,
            errors=errors,
            to_add={'code': 'empty'},
          )

      if self.max_length is not None:
        if len(value) > self.max_length:
          self._append_error(
            key=key,
            errors=errors,
            to_add={
              'code': 'maxLength',
              'expected': self.max_length,
              'received': len(value),
            },
          )

      if self.min_length is not None:
        if len(value) < self.min_length:
          self._append_error(
            key=key,
            errors=errors,
            to_add={
              'code': 'minLength',
              'expected': self.min_length,
              'received': len(value),
            },
          )

      if self.choices is not None:
        mapped_choices = [choice[0] for choice in self.choices]
        if value not in mapped_choices:
          self._append_error(
            key=key,
            errors=errors,
            to_add={
              'code': 'invalidChoice',
              'expected': mapped_choices,
              'received': value,
            },
          )



""" Base class for the fields """


class Field:
  """ Field abstract class """

  def __init__(self, required=False):
    self.required = required

  def validate(self, key, value, errors):
    """ Validate is the field is blank or None if is required
    ---
    Arguments
      key: str
        Key of the field
      value: any
        Value to validate
      errors: dict
        Dict of errors
    """

    if self.required:
      if value is None:
        self._append_error(key=key, errors=errors, to_add={'code': 'required'})

  def _convert_to_camel(self, key):
    """
    Convert the key to camel case
    """
    init, *temp = key.split('_')

    field = ''.join([init, *map(str.title, temp)])
    field_items = field.split(".")

    field_final = []
    for item in field_items:
      field_final.append(''.join([item[0].lower(), item[1:]]))

    return '.'.join(field_final)

  def _append_error(self, key, errors, to_add):
    """
    Append an error to a dict of errors
    ---
    Arguments
      key: str
        Key of the error
      errors: dict
        Dict of errors
      to_add: dict
        Error to add
    """

    key = self._convert_to_camel(key=key)
    if key in errors:
      errors[key].append(to_add)
    else:
      errors[key] = [to_add]


""" Boolean field """
from .base import Field


class BooleanField(Field):
  """
  IdField class for validation
  ---
  Attributes
    required: bool
      Indicates if the field is required or not
  """

  def __init__(self, required=False):
    super(BooleanField, self).__init__(required=required)

  def validate(self, key, value, errors):
    """
    Validate the field with the following rules:
    - Should be a bool
    ---
    Arguments
      key: str
        Key of the field
      value: any
        Value to validate
      errors: dict
        Dict of errors
    """

    super(BooleanField, self).validate(key=key, value=value, errors=errors)

    if not isinstance(value, bool) and (self.required and value is not None):
      self._append_error(
        key=key,
        errors=errors,
        to_add={'code': 'invalid'},
      )



""" Email field """
import re
from .base import Field


class EmailField(Field):
  """
  EmailField class for validation
  ---
  Attributes
    required: bool
      Indicates if the field is required or not
    empty: bool
      Indicates if the field can be empty or not
    regex: str
      Regex to validate the email
  """

  def __init__(
    self,
    required=False,
    empty=False,
    regex=r'^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-z]{2,63}$',
  ):
    super(EmailField, self).__init__(required=required)
    self.empty = empty
    self.regex = regex

  def validate(self, key, value, errors):
    """
    Validate the field with the following rules:
    - Should be a valid email, the validation will compile the regex
    ---
    Arguments
      key: str
        Key of the field
      value: any
        Value to validate
      errors: dict
        Dict of errors
    """

    super(EmailField, self).validate(key=key, value=value, errors=errors)

    if isinstance(value, str):
      if not self.empty:
        if value == '' or value is None:
          self._append_error(
            key=key,
            errors=errors,
            to_add={'code': 'required'},
          )
        else:
          if not re.match(self.regex, value):
            self._append_error(
              key=key,
              errors=errors,
              to_add={'code': 'invalid'},
            )
    else:
      self._append_error(
        key=key,
        errors=errors,
        to_add={'code': 'invalid'},
      )



""" JSON Field """
from .base import Field


class JsonField(Field):
  """
  JsonField class for validation
  ---
  Attributes
    required: bool
      Indicates if the field is required or not
    empty: bool
      Indicates if the field can be empty or not
    datatype: (dict, list)
      Indicates the datatype of the field
  """

  def __init__(
    self,
    required=False,
    empty=False,
    datatype=dict,
  ):
    super(JsonField, self).__init__(required=required)
    self.empty = empty
    self.datatype = datatype

  def validate(self, key, value, errors):
    """
    Validate the field with the following rules:
    - Should be a dict or list (Depending of the datatype)
    - If `empty` is False, the field should not be empty
      * For `dict`, should have at least 1 key
      * For `list`, should have at least 1 item
    ---
    Arguments
      key: str
        Key of the field
      value: any
        Value to validate
      errors: dict
        Dict of errors
    """

    super(JsonField, self).validate(key=key, value=value, errors=errors)

    if not isinstance(value, self.datatype):
      self._append_error(
        key=key,
        errors=errors,
        to_add={'code': 'invalid'},
      )
    elif not self.empty:
      length = 0

      if isinstance(self.datatype(), dict):
        length = len(value.keys())
      elif isinstance(self.datatype(), list):
        length = len(value)

      if length == 0:
        self._append_error(
          key=key,
          errors=errors,
          to_add={'code': 'invalid'},
        )
